#include <profiler>
#pragma newdecls required

public Plugin myinfo =
{
	name = "sourcemod.xml generator",
	author = "MCPAN (mcpan@foxmail.com)",
	version = "1.1.0",
	url = "https://forums.alliedmods.net/member.php?u=73370"
}

#define	MAX_WIDTH 28
#define WIDTH MAX_WIDTH - 4

#define SPACE_CHAR	' '
#define SPACE_X4	"    "
#define SPACE_X8	"        "
#define SPACE_X12	"            "
#define SPACE_X16	"                "
#define SPACE_X28	"                            "

#define COMMENT_PARAM		"@param"
#define COMMENT_RETURN		"@return"
#define COMMENT_NORETURN	"@noreturn"
#define COMMENT_ERROR		"@error"

#define PATH_INCLUDE	"addons/sourcemod/scripting/include"
#define FILE_SOURCEMOD	"addons/sourcemod/plugins/sourcemod.xml"
#define FILE_FUNCTIONS	"addons/sourcemod/plugins/all_function.sp"
#define FILE_DEFINES	"addons/sourcemod/plugins/all_define.sp"

StringMap g_FuncTrie;
ArrayList g_FuncArray;
ArrayList g_DefineArray;
File g_FileSourcemod;

public void OnPluginStart()
{
	RegServerCmd("test", Cmd_Start);
}

public Action Cmd_Start(int argc)
{
	Handle prof = CreateProfiler();
	StartProfiling(prof);
	
	int size;
	char buffer[PLATFORM_MAX_PATH];
	ArrayList fileArray = CreateArray(ByteCountToCells(PLATFORM_MAX_PATH));
	
	g_FuncTrie = CreateTrie();
	g_FuncArray = CreateArray(ByteCountToCells(64));
	g_DefineArray = CreateArray(ByteCountToCells(64));
	
	if ((size = ReadDirFileList(fileArray, PATH_INCLUDE, "inc")))
	{
		for (int i = 0; i < size; i++)
		{
			fileArray.GetString(i, buffer, PLATFORM_MAX_PATH-1);
			ReadIncludeFile(buffer, i);
		}
	}
	
	SortADTArrayCustom(g_FuncArray, SortFuncADTArray);
	SortADTArrayCustom(g_DefineArray, SortFuncADTArray);
	
	File file = OpenFile(FILE_FUNCTIONS, "wb");
	g_FileSourcemod = OpenFile(FILE_SOURCEMOD, "wb");
	
	g_FileSourcemod.WriteLine("<?xml version=\"1.0\" encoding=\"Windows-1252\" ?>");
	g_FileSourcemod.WriteLine("<NotepadPlus>");
	g_FileSourcemod.WriteLine("%s<AutoComplete language=\"sourcemod\">", SPACE_X4);
	
	if ((size = GetArraySize(g_FuncArray)))
	{
		int value;
		char funcname[64];
		for (int i = 0; i < size; i++)
		{
			g_FuncArray.GetString(i, funcname, 63);
			g_FuncTrie.GetValue(funcname, value)
			fileArray.GetString(value, buffer, PLATFORM_MAX_PATH-1);
			ReadIncludeFile(buffer, _, funcname);
			file.WriteLine(funcname);
		}
	}
	
	g_FileSourcemod.WriteLine("%s</AutoComplete>", SPACE_X4);
	g_FileSourcemod.WriteLine("</NotepadPlus>");
	
	delete file;
	delete fileArray;
	delete g_FuncTrie;
	delete g_FuncArray;
	delete g_FileSourcemod;
	
	file = OpenFile(FILE_DEFINES, "wb");
	if ((size = GetArraySize(g_DefineArray)))
	{
		for (int i = 0; i < size; i++)
		{
			g_DefineArray.GetString(i, buffer, PLATFORM_MAX_PATH-1);
			file.WriteLine(buffer);
		}
	}
	
	delete file;
	delete g_DefineArray;
	
	StopProfiling(prof);
	PrintToServer("\n\t\t\t\tDone. time used %fs", GetProfilerTime(prof));
	delete prof;
	
	return Plugin_Handled;
}

void ReadIncludeFile(char[] filepath, int fileArrayIdx=-1, char[] search="")
{
	File file;
	if ((file = OpenFile(filepath, "rb")) == INVALID_HANDLE)
	{
		LogError("Open file faild '%s'", filepath);
		return;
	}
	
	int value, i;
	bool comment_buffer, found_comment, found_params, found_return, found_error, found_func;
	char temp[1024], buffer[1024], funcprefix[8], retval[32], funcname[64], funcparam[32];
	ArrayList array_param, array_return, array_error, array_note
	
	array_param = CreateArray(ByteCountToCells(1024));
	array_return = CreateArray(ByteCountToCells(1024));
	array_error = CreateArray(ByteCountToCells(1024));
	array_note = CreateArray(ByteCountToCells(1024));

	while (ReadFileLine(file, buffer, 1023))
	{
		if (!ReadString(buffer, 1023, found_comment))
		{
			if (found_comment)
			{
				found_params = false;
				found_return = false;
				found_error = false;
				ClearArray(array_param);
				ClearArray(array_return);
				ClearArray(array_error);
				ClearArray(array_note);
			}
			continue;
		}
		
		if (found_comment)
		{
			if (!search[0])
			{
				continue;
			}
			
			if ((value = FindCharInString2(buffer, '*')) != -1)
			{
				strcopy(buffer, 1023, buffer[++value]);
			}
			
			TrimString(buffer);
			
			if (!buffer[0])
			{
				continue;
			}
			
			if (StrContains(buffer, COMMENT_PARAM) == -1 &&
				StrContains(buffer, COMMENT_RETURN) == -1 &&
				StrContains(buffer, COMMENT_NORETURN) == -1 &&
				StrContains(buffer, COMMENT_ERROR) == -1)
			{
				if (found_params)
				{
					FormatEx(temp, 1023, "%s%s", SPACE_X28, buffer);
					PushArrayString(array_param, temp);
				}
				else if (found_return)
				{
					FormatEx(temp, 1023, "%s%s", SPACE_X4, buffer);
					PushArrayString(array_return, temp);
				}
				else if (found_error)
				{
					FormatEx(temp, 1023, "%s%s", SPACE_X4, buffer);
					PushArrayString(array_error, temp);
				}
				else
				{
					ReplaceString(buffer, 1023, "@note", "");
					ReplaceString(buffer, 1023, "@brief", "");
					
					TrimString(buffer);
					FormatEx(temp, 1023, "%s%s", SPACE_X4, buffer);
					PushArrayString(array_note, temp);
				}
			}
			else if ((value = StrContains(buffer, COMMENT_PARAM)) != -1)
			{
				found_params = true;
				found_return = false;
				found_error = false;
				strcopy(buffer, 1023, buffer[value+6]);
				TrimString(buffer);
				
				if (buffer[0] && (value = FindCharInString2(buffer, SPACE_CHAR)) != -1)
				{
					strcopy(funcparam, value+1, buffer);
					strcopy(buffer, 1023, buffer[value]);
					TrimString(buffer);
					
					if ((value = WIDTH - value) > 0)
					{
						for (i = 0; i < value; i++)
						{
							temp[i] = SPACE_CHAR;
						}
						temp[value] = 0;
					}
					else
					{
						LogMessage("need space, set MAX_WIDTH >= %d", MAX_WIDTH - value);
					}
					
					Format(temp, 1023, "%s%s%s%s", SPACE_X4, funcparam, value > 0 ? temp : SPACE_X4, buffer);
					PushArrayString(array_param, temp);
				}
			}
			else if ((value = StrContains(buffer, COMMENT_RETURN)) != -1 || StrContains(buffer, COMMENT_NORETURN) != -1)
			{
				found_params = false;
				found_return = true;
				found_error = false;
				
				if (StrContains(buffer, COMMENT_NORETURN) != -1)
				{
					found_return = false;
					continue;
				}
				
				strcopy(buffer, 1023, buffer[value+7]);
				TrimString(buffer);
				FormatEx(temp, 1023, "%s%s", SPACE_X4, buffer);
				PushArrayString(array_return, temp);
			}
			else if ((value = StrContains(buffer, COMMENT_ERROR)) != -1)
			{
				found_params = false;
				found_return = false;
				found_error = true;
				strcopy(buffer, 1023, buffer[value+6]);
				TrimString(buffer);
				FormatEx(temp, 1023, "%s%s", SPACE_X4, buffer);
				PushArrayString(array_error, temp);
			}
			else
			{
				LogMessage(buffer);
			}
		}
		else if (StrContains(buffer, "#pragma deprecated") != -1 && ReadFileLine(file, buffer, 1023))
		{
			strcopy(funcprefix, 7, buffer);
			TrimString(funcprefix);
			
			do
			{
				if (strcmp(funcprefix, "stock") == 0 && buffer[0] == '}' ||
					strcmp(funcprefix, "stock") != 0 && FindCharInString2(buffer, ')') != -1)
				{
					break;
				}
			}
			while (ReadFileLine(file, buffer, 1023));
		}
		else
		{
			if ((value = StrContains(buffer, "#define ")) != -1)
			{
				if (search[0] ||
					StrContains(buffer, "_included") != -1 ||
					FindCharInString2(buffer, '(') != -1 ||
					FindCharInString2(buffer, '[') != -1)
				{
					continue;
				}
				
				strcopy(buffer, 1023, buffer[value+7]);
				TrimString(buffer);
				
				if ((value = FindCharInString2(buffer, SPACE_CHAR)) != -1)
				{
					strcopy(buffer, ++value, buffer);
					TrimString(buffer);
				}
				
				if (IsValidString(buffer) && FindStringInArray(g_DefineArray, buffer) == -1)
				{
					PushArrayString(g_DefineArray, buffer);
				}
			}
			else if ((value = StrContains(buffer, "enum")) != -1)
			{
				if (search[0])
				{
					continue;
				}
				
				strcopy(buffer, 1023, buffer[value+4]);
				TrimString(buffer);
				
				if (value)
				{
					if (IsValidString(buffer) && FindStringInArray(g_DefineArray, buffer) == -1)
					{
						PushArrayString(g_DefineArray, buffer);
					}
					continue;
				}
				
				if ((value = FindCharInString2(buffer, '{')) != -1)
				{
					strcopy(temp, ++value, buffer);
					strcopy(buffer, 1023, buffer[value]);
					TrimString(temp);
					TrimString(buffer);
					
					if (WriteDefines(g_DefineArray, buffer, 1023, FindCharInString2(buffer, '}')))
					{
						while (ReadFileLine(file, buffer, 1023))
						{
							if (!WriteDefines(g_DefineArray, buffer, 1023, FindCharInString2(buffer, '}')))
							{
								break;
							}
						}
					}
				}
				else if (IsValidString(buffer) && FindStringInArray(g_DefineArray, buffer) == -1)
				{
					PushArrayString(g_DefineArray, buffer);
					
					while (ReadFileLine(file, buffer, 1023))
					{
						if (!ReadString(buffer, 1023, found_comment, comment_buffer) || comment_buffer)
						{
							continue;
						}
						
						if ((value = FindCharInString2(buffer, '{')) != -1)
						{
							strcopy(temp, ++value, buffer);
							strcopy(buffer, 1023, buffer[value]);
							TrimString(temp);
							TrimString(buffer);
							
							if (WriteDefines(g_DefineArray, buffer, 1023, FindCharInString2(buffer, '}')))
							{
								while (ReadFileLine(file, buffer, 1023))
								{
									if (!ReadString(buffer, 1023, found_comment, comment_buffer) || comment_buffer)
									{
										continue;
									}
									
									if (!WriteDefines(g_DefineArray, buffer, 1023, FindCharInString2(buffer, '}')))
									{
										break;
									}
								}
							}
							break;
						}
					}
				}
			}
			else
			{
				strcopy(funcprefix, 7, buffer);
				TrimString(funcprefix);
				
				found_func = false;
				if (strcmp(funcprefix, "forward") == 0)
				{
					found_func = true;
					strcopy(buffer, 1015, buffer[8]);
				}
				else if (strcmp(funcprefix, "native") == 0)
				{
					found_func = true;
					strcopy(buffer, 1016, buffer[7]);
				}
				else if (strcmp(funcprefix, "stock") == 0)
				{
					found_func = true;
					strcopy(buffer, 1017, buffer[6]);
				}
				
				if (found_func && ReadFuncString(buffer, retval, funcname) && IsValidString(funcname))
				{
					if (search[0])
					{
						if (strcmp(funcname, search) == 0)
						{
							WriteFileLine(g_FileSourcemod, "%s<KeyWord name=\"%s\" func=\"yes\">", SPACE_X8, funcname);
							WriteFileLine(g_FileSourcemod, "%s<Overload retVal=\"%s %s\" descr=\"", SPACE_X12, funcprefix, retval[0] ? retval : "void");
							
							if ((value = GetArraySize(array_param)))
							{
								WriteFileLine(g_FileSourcemod, "Params:");
								for (i = 0; i < value; i++)
								{
									temp[0] = 0;
									GetArrayString(array_param, i, temp, 1023);
									WriteFileLine(g_FileSourcemod, temp);
								}
							}
							if ((value = GetArraySize(array_note)))
							{
								WriteFileLine(g_FileSourcemod, "Notes:");
								for (i = 0; i < value; i++)
								{
									temp[0] = 0;
									GetArrayString(array_note, i, temp, 1023);
									WriteFileLine(g_FileSourcemod, temp);
								}
							}
							if ((value = GetArraySize(array_error)))
							{
								WriteFileLine(g_FileSourcemod, "Error:");
								for (i = 0; i < value; i++)
								{
									temp[0] = 0;
									GetArrayString(array_error, i, temp, 1023);
									WriteFileLine(g_FileSourcemod, temp);
								}
							}
							if ((value = GetArraySize(array_return)))
							{
								WriteFileLine(g_FileSourcemod, "Return:");
								for (i = 0; i < value; i++)
								{
									temp[0] = 0;
									GetArrayString(array_return, i, temp, 1023);
									WriteFileLine(g_FileSourcemod, temp);
								}
							}
							
							WriteFileLine(g_FileSourcemod, "\">");
							
							if (buffer[0] == '(')
							{
								buffer[0] = SPACE_CHAR;
							}
							
							if (WriteFuncParams(g_FileSourcemod, buffer, 1023, FindCharInString2(buffer, ')')))
							{
								while (ReadFileLine(file, buffer, 1023))
								{
									if (!WriteFuncParams(g_FileSourcemod, buffer, 1023, FindCharInString2(buffer, ')')))
									{
										break;
									}
								}
							}
							
							WriteFileLine(g_FileSourcemod, "%s</Overload>", SPACE_X12);
							WriteFileLine(g_FileSourcemod, "%s</KeyWord>", SPACE_X8);
						}
						
						ClearArray(array_param);
						ClearArray(array_return);
						ClearArray(array_error);
						ClearArray(array_note);
					}
					else if (FindStringInArray(g_FuncArray, funcname) == -1)
					{
						PushArrayString(g_FuncArray, funcname);
						SetTrieValue(g_FuncTrie, funcname, fileArrayIdx);
					}
				}
			}
		}
	}
	
	CloseHandle(array_param);
	CloseHandle(array_return);
	CloseHandle(array_error);
	CloseHandle(array_note);
	CloseHandle(file);
}

int ReadString(char[] buffer, int maxlength, bool &found_comment=false, bool &comment_buffer=false)
{
	ReplaceString(buffer, maxlength, "\t", " ");
	ReplaceString(buffer, maxlength, "\"", "'");
	ReplaceString(buffer, maxlength, "%", "%%");
	
	static int len, i;
	if ((len = strlen(buffer)) && !found_comment)
	{
		for (i = 0; i < len; i++)
		{
			if (buffer[i] == '/' && buffer[i+1] == '/')
			{
				buffer[i] = 0;
				break;
			}
		}
	}
	
	static bool comment_start, comment_end;
	static int pos
	static char temp[1024]
	pos = 0
	comment_start = comment_end = false
	
	TrimString(buffer);
	if ((len = strlen(buffer)))
	{
		if (found_comment)
			comment_buffer = true;

		if ((pos = StrContains(buffer, "/*")) != -1)
		{
			comment_start = true;
			strcopy(temp, 1023, buffer[pos+2]);
			buffer[pos] = 0;
			TrimString(buffer);
			
			if ((pos = StrContains(temp, "*/")) != -1)
			{
				comment_end = true;
				strcopy(temp, 1023, temp[pos+2]);
				TrimString(temp);
			}
			else
			{
				temp[0] = 0;
			}
			
			if (strlen(buffer) || strlen(temp))
			{
				comment_buffer = false;
				Format(buffer, maxlength, "%s%s", buffer, temp);
			}
			temp[0] = 0;
		}
		else if ((pos = StrContains(buffer, "*/")) != -1)
		{
			comment_end = true;
			comment_buffer = false;
			strcopy(buffer, maxlength, buffer[pos+2]);
		}
		
		TrimString(buffer);
		len = strlen(buffer);
	}
	
	if (comment_start && comment_end)
	{
		comment_start = false;
		comment_end = false;
	}
	
	if (comment_start)
	{
		found_comment = comment_start;
	}
	else if (comment_end)
	{
		found_comment = comment_start;
	}
	
	return len;
}

bool ReadFuncString(char[] buffer, char[] retval, char[] funcname)
{
	retval[0] = 0;
	funcname[0] = 0;
	
	static int pos, len;
	if ((len = strlen(buffer)) && (pos = FindCharInString2(buffer, '(')) != -1)
	{
		strcopy(funcname, pos+1, buffer);
		strcopy(buffer, len, buffer[pos]);
		
		if (strcmp(funcname, "VerifyCoreVersion") == 0 ||
			strcmp(funcname, "operator%%") == 0 ||
			StrContains(funcname, ":operator") != -1)
		{
			return false;
		}
		
		if ((pos = FindCharInString2(funcname, 32)) != -1 || (pos = FindCharInString2(funcname, 58)) != -1)
		{
			strcopy(retval, ++pos, funcname);
			strcopy(funcname, len, funcname[pos]);

			if(retval[0] == 70) // little fix 'F' -> 'f'
				retval[0] = 102
		}
		
		return true;
	}
	
	return false;
}

bool WriteFuncParams(Handle handle, char[] buffer, int maxlength, int pos)
{
	if (pos != -1)
	{
		buffer[pos] = 0;
	}
	
	ReplaceString(buffer, maxlength, "\t", " ");
	ReplaceString(buffer, maxlength, "\"", "'");
	ReplaceString(buffer, maxlength, "%", "%%");
	
	TrimString(buffer);
	if (buffer[0])
	{
		static char funcparams[32][256];
		static int count, i
		count = ExplodeString(buffer, ",", funcparams, sizeof(funcparams), sizeof(funcparams[]));
		for (i = 0; i < count; i++)
		{
			TrimString(funcparams[i]);
			if (funcparams[i][0])
			{
				WriteFileLine(handle, "%s<Param name=\"%s\"/>", SPACE_X16, funcparams[i]);
				funcparams[i][0] = 0;
			}
		}
	}
	
	return pos == -1;
}

bool WriteDefines(Handle &handle, char[] buffer, int maxlength, int pos)
{
	if (pos != -1)
	{
		buffer[pos] = 0;
	}
	
	ReplaceString(buffer, maxlength, "\t", " ");
	ReplaceString(buffer, maxlength, "\"", "'");
	
	TrimString(buffer);
	if (buffer[0])
	{
		static char defines_temp[32][64];
		int pos2, i, value
		value = ExplodeString(buffer, ",", defines_temp, sizeof(defines_temp), sizeof(defines_temp[]));
		for (i = 0; i < value; i++)
		{
			TrimString(defines_temp[i]);
			if (defines_temp[i][0])
			{
				if ((pos2 = FindCharInString2(defines_temp[i], '=')) != -1)
				{
					defines_temp[i][pos2] = 0;
					TrimString(defines_temp[i]);
				}
				
				if (IsValidString(defines_temp[i]) && FindStringInArray(handle, defines_temp[i]) == -1)
				{
					PushArrayString(handle, defines_temp[i]);
				}
				
				defines_temp[i][0] = 0;
			}
		}
	}
	
	return pos == -1;
}

bool IsValidString(char[] buffer)
{
	TrimString(buffer);
	return (buffer[0] &&
			FindCharInString2(buffer, SPACE_CHAR) == -1 &&
			FindCharInString2(buffer, '*') == -1 &&
			FindCharInString2(buffer, '/') == -1 &&
			FindCharInString2(buffer, ':') == -1 &&
			FindCharInString2(buffer, '(') == -1 &&
			FindCharInString2(buffer, '[') == -1 &&
			FindCharInString2(buffer, ']') == -1 &&
			FindCharInString2(buffer, ')') == -1 &&
			FindCharInString2(buffer, '%') == -1 &&
			StrContains(buffer, "__FLOAT") == -1);
}

public int SortFuncADTArray(int index1, int index2, Handle array, Handle hndl)
{
	char str1[64], str2[64];
	GetArrayString(array, index1, str1, 63);
	GetArrayString(array, index2, str2, 63);
	return strcmp(str1, str2, false);
}

stock int ReadDirFileList(Handle &fileArray, const char[] dirPath, const char[] fileExt="")
{
	Handle dir;
	if ((dir = OpenDirectory(dirPath)) == INVALID_HANDLE)
	{
		LogError("Open dir faild '%s'", dirPath);
		return 0;
	}
	
	FileType fileType;
	char buffer[PLATFORM_MAX_PATH], currentPath[PLATFORM_MAX_PATH];
	ArrayList pathArray = CreateArray(ByteCountToCells(PLATFORM_MAX_PATH));
	
	buffer[0] = 0;
	currentPath[0] = 0;
	
	while (ReadDirEntry(dir, buffer, PLATFORM_MAX_PATH-1, fileType)
		|| ReadSubDirEntry(dir, buffer, PLATFORM_MAX_PATH-1, fileType, pathArray, dirPath, currentPath))
	{
		switch (fileType)
		{
			case FileType_Directory:
			{
				if (strcmp(buffer, ".") != 0 && strcmp(buffer, "..") != 0)
				{
					Format(buffer, PLATFORM_MAX_PATH-1, "%s/%s", currentPath, buffer);
					PushArrayString(pathArray, buffer);
				}
			}
			case FileType_File:
			{
				if (fileExt[0] && !CheckFileExt(buffer, fileExt))
				{
					continue;
				}
				
				Format(buffer, PLATFORM_MAX_PATH-1, "%s%s/%s", dirPath, currentPath, buffer);
				PushArrayString(fileArray, buffer);
			}
		}
	}
	
	CloseHandle(pathArray);
	if (dir != INVALID_HANDLE)
	{
		CloseHandle(dir);
	}
	
	return GetArraySize(fileArray);
}

stock bool ReadSubDirEntry(Handle &dir, char[] buffer, int maxlength, FileType &fileType, Handle &pathArray, const char[] dirPath, char[] currentPath)
{
	if (!GetArraySize(pathArray))
	{
		return false;
	}
	
	GetArrayString(pathArray, 0, currentPath, maxlength);
	RemoveFromArray(pathArray, 0);
	
	CloseHandle(dir);
	dir = INVALID_HANDLE;
	
	FormatEx(buffer, maxlength, "%s%s", dirPath, currentPath);
	if ((dir = OpenDirectory(buffer)) == INVALID_HANDLE)
	{
		LogError("Open sub dir faild '%s'", buffer);
		return false;
	}
	
	return ReadDirEntry(dir, buffer, maxlength, fileType);
}

stock bool CheckFileExt(char[] filename, const char[] extname)
{
	int pos;
	if ((pos = FindCharInString2(filename, '.', true)) == -1)
	{
		return false;
	}
	
	char ext[32];
	strcopy(ext, 31, filename[++pos]);
	return strcmp(ext, extname, false) == 0;
}

int FindCharInString2(const char[] str, char c, bool reverse = false)
{
	static int len, i
	len = strlen(str);
	
	if (!reverse) {
		for (i = 0; i < len; i++) {
			if (str[i] == c)
				return i;
		}
	} else {
		for (i = len - 1; i >= 0; i--) {
			if (str[i] == c)
				return i;
		}
	}

	return -1;
}